/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"
#include "fftw/fftw3.h"
#include <string.h>

extern ULONG DFTmax,DFToptimize;
extern int DFTdmin,DFTdmax,DFTdstp, // DFT bit distance
	DFTmmin,DFTmstp; // DFT mode

#define CONCAT(prefix, name) prefix ## name

	#if 0
#define TYPE double
#define X(name) CONCAT(fftw_, name)
#define WISDOM_name "DFTwisdom_d"
	#else
#define TYPE float
#define X(name) CONCAT(fftwf_, name)
#define WISDOM_name "DFTwisdom_f"
	#endif

static X(plan) *pln=NULL; // Pointer to the planes for each bit distance (DFTdxxx)
static int plnC=0; // Number of planes used
static TYPE *in=NULL;
static X(complex) *out=NULL;
static double *xr=NULL; // Normalized out[][1] for the KS test
static ULONG N; // n used for the DFT (truncated and optimized)

void DFTfree(void)
{
	for(int p=0;p<plnC;p++) X(destroy_plan)(pln[p]);
	delete[] pln; fftw_free(in); fftw_free(out); delete[] xr;
}

void DFTinit(void)
{
	N=n;
	if(DFTmax) // Truncate n to save running time and memory
		if(DFTmax<n) {
			N=DFTmax; printf("\n  Sequence length for the DFT test truncated to DFTmax (%u bits)\n\n",N);
		}

	if(N>DFToptimize) { // Find the best N (see fftw3.pdf page 24)
		ULONG NN=N,DM=0,TM=0,CM=0,SM=0,err=N,d,t,c,s,n11,n13;

		t=N; while(t>=2) { t/=2; DM++; } // DM=log_2(N)
		t=N; while(t>=3) { t/=3; TM++; } // TM=log_3(N)
		t=N; while(t>=5) { t/=5; CM++; } // CM=log_5(N)
		t=N; while(t>=7) { t/=7; SM++; } // SM=log_7(N)

		for(int T=0;T<=TM;T++) { ULONG TT=pow(3.,T);
			for(int C=0;C<=CM;C++) { ULONG CC=pow(5.,C);
				for(int S=0;S<=SM;S++) { ULONG SS=pow(7.,S);
					for(int N11=0; N11<=1; N11++) {
						for(int N13=0; N13<=1; N13++) {
							if(N11 && N13) continue;
							for(int D=0;D<=DM;D++) {
								UI64 n=TT*CC*SS*(UI64(1)<<D)*(N11 ? 11:1)*(N13 ? 13:1);
								if(n>N) break;
								if(N-n<err) { err=N-n; d=D; t=T; c=C; s=S; n11=N11; n13=N13; }
							}
						}
					}
				}
			}
		}

		N=pow(2.,double(d))*pow(3.,double(t))*pow(5.,double(c))*pow(7.,double(s))*
			(n11 ? 11:1)*(n13 ? 13:1);
		
		if(N!=NN) printf("\n  Sequence length for the DFT test optimized to %u bits (%d bits)\n\n",N,N-NN);
	}

	// LOAD WISDOM
	char *Ws; FILE *Wf=fopen(WISDOM_name,"rb"); if(Wf) {
		int l; if(sizeof(int)==fread(&l,1,sizeof(int),Wf)) {
			Ws=new char[l+1]; Ws[l]=0;
			if(l==fread(Ws,1,l,Wf)) {
				if(0==X(import_wisdom_from_string)(Ws)) puts(" WARNING: wisdom not imported (DFT test).");
			} else puts(" WARNING: wisdom not loaded (DFT test).");
			delete[] Ws;
		}
		fclose(Wf);
	}

	// Count how many plans will be used (1 plane for each nx)
	ULONG nxy=0,nxy2=0,ny2m=0; plnC=0; int nx;
	for(nx=DFTdmin; nx<=DFTdmax; INCR(nx,DFTdstp)) {
		plnC++; ULONG ny=N/nx,ny2=ny/2+1; // Find the biggest vector size
		nxy=max(nxy,nx*ny); nxy2=max(nxy2,nx*ny2); ny2m=max(ny2m,ny2);
	}
	
	// Allocate the memory for the plans, in, out e KS vectors
	in= (TYPE*) fftw_malloc(sizeof(TYPE) * nxy);
	out=(X(complex)*) fftw_malloc(sizeof(X(complex)) * nxy2);
//	xr=new double[ny2m]; pln=new X(plan)[plnC]; //printf("%u kbytes allocated\n",(nxy*sizeof(TYPE) + nxy2*sizeof(X(complex)) + ny2m*sizeof(double))/1024);
	xr=new double[nxy2]; pln=new X(plan)[plnC]; //printf("%u kbytes allocated\n",(nxy*sizeof(TYPE) + nxy2*sizeof(X(complex)) + nxy2*sizeof(double))/1024);

	// Initialize the plans which will be used for each nx
	plnC=0;
	for(nx=DFTdmin; nx<=DFTdmax; INCR(nx,DFTdstp))
		pln[plnC++]= X(plan_dft_r2c_2d)(nx,N/nx, in,out, FFTW_ESTIMATE | FFTW_DESTROY_INPUT);

	// SAVE WISDOM
	Wf=fopen(WISDOM_name,"wb"); if(Wf) {
		Ws=X(export_wisdom_to_string)(); int l=strlen(Ws);
		if(sizeof(int)==fwrite(&l,1,sizeof(int),Wf)) {
			if(l!=fwrite(Ws,1,l,Wf)) puts(" WARNING: wisdom not saved (DFT test).");
		}
		fftw_free(Ws); fclose(Wf);
	}
}


void DFT(void)
{
	static int flg=1; if(flg) { DFTinit(); flg=0; } // Initialization (N, in and out MUST NOT CHANGE)

	for(int nx=DFTdmin,p=0; nx<=DFTdmax; INCR(nx,DFTdstp),p++) { // Rows

		const ULONG ny=N/nx,ny2=ny/2+1;
		
		if(ny2<1000) break; // At least 1000 bits in each test

		const double SD= sqrt(nx/8.*ny); // Expected standard deviation

		for(int mode=DFTmmin; mode<2; mode+=DFTmstp) { // Mode (0, 1)

			ULONG C=0;
			if(nx==1 || mode==0) {
				for(;C<nx*ny;C++) in[C]=get_bit(C,bits);
				if(nx==1) mode=1; // With 1-D DFT, mode0 = mode1
			}
			else // MODE 1
				for(ULONG x=0;x<nx;x++)
					for(ULONG y=0;y<ny;y++) in[C++]=get_bit(x+y*nx,bits);

			X(execute)(pln[p]); // Calculate the DFT using the appropriate plane

			// KS test over the imaginary part (normal distribution)
			C=0; for(int x=0;x<nx;x++) for(int y=0;y<ny2;y++) xr[C++]=out[(y+ny2*x)][1]/SD;
			if(nx==1) strcpy(buf,"DFT 1"); else sprintf(buf,"DFT %d_%d",nx,mode);
			SAVE(buf,KSnorm(xr,C));

			/*for(int x=0;x<nx;x++) { // This generates many p-vals; the results must be interpreted carefully
				for(int y=0;y<ny2;y++) xr[y]=out[(y+ny2*x)][1]/SD;
				sprintf(buf,"DFT %d_%d_%d",nx,mode,x); SAVE(buf,KSnorm(xr,ny2));
			}*/

		} // mode

	} // nx
}
