/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

/*
This is the BINARY RANK TEST for 6x8 matrices.  From each of 
six random 32-bit integers from the generator under test, a  
specified byte is chosen, and the resulting six bytes form a 
6x8 binary matrix whose rank is determined.  That rank can be
from 0 to 6, but ranks 0,1,2,3 are rare; their counts are    
pooled with those for rank 4. Ranks are found for 100,000    
random matrices, and a chi-square test is performed on       
counts for ranks <=4, 5 and 6.

This is the BINARY RANK TEST for 31x31 matrices. The leftmost
31 bits of 31 random integers from the test sequence are used
to form a 31x31 binary matrix over the field {0,1}. The rank 
is determined. That rank can be from 0 to 31, but ranks< 28  
are rare, and their counts are pooled with those for rank 28.
Ranks are found for 40,000 such random matrices and a chisqu-
are test is performed on counts for ranks 31,30,28 and <=28. 
 (The 31x31 choice is based on the unjustified popularity of 
  the proposed "industry standard" generator
  x(n) = 16807*x(n-1) mod 2^31-1, not a very good one.)      

This is the BINARY RANK TEST for 32x32 matrices. A random
32x32 binary matrix is formed, each row a 32-bit random integer.
The rank is determined. That rank can be from 0 to 32. Ranks 
less than 29 are rare, and their counts are pooled with those
for rank 29.  Ranks are found for 40,000 such random matrices
and a chisquare test is performed on counts for ranks  32,31,
30 and <=29.
*/

#include "defi.h"
#include <malloc.h>
#include <string.h>

// define a binary matrix
struct binmatrix {
    ULONG *row,mask;
    int no_row, no_col;
}; 

extern int RNKmin,RNKmax,RNKstp;

// compute the rank of a binary matrix
static int _FASTCALL_ rkbm(binmatrix bm)
{
    register int i, j, k, rt=0;
    int rank=0;
    ULONG tmp;

    for(k=0; k<bm.no_row; ++k) {
        i=k;

        while( ( (bm.row[i]>>rt)&1 ) == 0 ) {
            ++i;

            if( i<bm.no_row ) continue;
            else {
                ++rt;
                if( rt<bm.no_col ) { i=k; continue; }
            }
            return rank;
        }

        ++rank;
        if( i!=k ){
            tmp=bm.row[i];
            bm.row[i]=bm.row[k];
            bm.row[k]=tmp;
        }

        for(j=i+1; j<bm.no_row; ++j)
            if( (bm.row[j] >> rt) & 1 ) bm.row[j] ^= bm.row[k];

        ++rt;
    }

    return rank;
}

// perform the test and calculate test-stat 
static double _FASTCALL_ rnk_stat(ULONG no_mtr, int test, int rt)
{
    double p6[]={.009443, 0.217439, 0.773118};
    double p30[]={.0052854502, .1283502644, .5775761902, .2887880952};

    binmatrix bm;
    register ULONG i, j;
    register int cls, llim;
    ULONG *f;
    ULONG df=3; double *p,Ef, chsq=0;

    switch(test) {
    case 1:
        df=2;
        bm.no_row=6; bm.no_col=8;
        bm.mask=255; // pow(2,8)-1;
        llim=4; p=p6;
        break;
    case 2:
        bm.no_row=bm.no_col=31;
        bm.mask=0x7fffffff; // pow(2,31)-1;
        llim=28; p=p30;
        break;
    case 3:
        bm.no_row=bm.no_col=32;
        bm.mask=0xffffffff; // 2^32 -1
        llim=29; p=p30;
        break;
    }

    bm.row=(ULONG*)malloc(bm.no_row*sizeof(ULONG));

    f=(ULONG*)calloc( (df+1), sizeof(ULONG));

    for( i=1; i<=no_mtr; ++i ){

        // get the rows of a matrix
        for(j=0; j<bm.no_row; ++j) {
            switch(test){
            case 1:
                bm.row[j]=rng()>>rt; break;
            case 2:
                bm.row[j]=rng()>>1; break;
            case 3:
                bm.row[j]=rng(); break;
            }
            bm.row[j] &= bm.mask;
        }

        cls=rkbm(bm); cls=max(llim,cls)-llim; ++f[cls];
    }

    // compute chi-square
    for(i=0; i<=df; ++i) {
        Ef=no_mtr* ( *(p+i) );
        chsq+=(f[i]-Ef)*(f[i]-Ef)/Ef;
    }

    free(bm.row);
    return chiQ( chsq,df );
}

static void _FASTCALL_ binrnk(const int test,const int mi,const int ma,const int stp)
{
    ULONG minrns=600000,no_matrices=100000;

    if(test>1) {
        no_matrices=40000;
        if (test==2) minrns = 1240000; // 31x31
        else minrns = 1280000; // 32x32
    }

    if(n < minrns*32) return;

	for(int rt=mi;rt<=ma;rt+=stp) {
		bpos=0; double p=rnk_stat(no_matrices, test, rt);

		if(test==1) sprintf(buf,"Rank 6x8 %d",rt);
		else if(test==2) strcpy(buf,"Rank 31x31");
		else strcpy(buf,"Rank 32x32");
		SAVE(buf,p);
    }
}


void RANK6_8  (void) { binrnk(1,RNKmin,RNKmax,RNKstp); }
void RANK31_31(void) { binrnk(2,0,0,1); }
void RANK32_32(void) { binrnk(3,0,0,1); }
